/*
    Tucnak - VHF contest log
    Copyright (C) 2002-2006  Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

struct dsp *dsp;

struct dsp *init_dsp(){
    struct dsp *dsp;

    dsp = g_new0(struct dsp, 1);   
    dsp->fd = -1;
    return dsp;
}

void free_dsp(struct dsp *dsp){
    if (!dsp) return;

    dsp_close(dsp);
	g_free(dsp);
}


int dsp_open(struct dsp *dsp, int rec){
    int tmp;

    CONDGFREE(dsp->filename);
    dsp->filename=g_strdup(cfg->ssbd_dsp);

/*    dbg("open_dsp(%d)\n", rec);*/
    dsp->fd = open (cfg->ssbd_dsp, rec?O_RDONLY:O_WRONLY, 0);
    if (dsp->fd < 0){
        log_addf("Can't open file '%s'", cfg->ssbd_dsp);
        goto err;
    }

   /* dsp->fragment=0x20009;
    dsp->fragment=0x3fff0009;
    if (ioctl (dsp->fd, SNDCTL_DSP_SETFRAGMENT, &dsp->fragment)){
        log_addf("Can't set fragment size to 0x%x\n", dsp->fragment);
        goto err;
    }   
        */
    if (ioctl(dsp->fd, SNDCTL_DSP_GETFMTS, &dsp->fmt_mask)){
        log_addf("Can't get supported audio formats");
        goto err;
    }
    
    tmp=dsp->format;
    if (ioctl (dsp->fd, SNDCTL_DSP_SETFMT, &dsp->format)){
        log_addf("Can't set audio format %d, recommended %d", tmp, dsp->format);
        
        tmp=dsp->format;
        if (ioctl (dsp->fd, SNDCTL_DSP_SETFMT, &dsp->format)){
            log_addf("Can't set recommended audio format %d, return %d", tmp, dsp->format);
            
            tmp = dsp->format = AFMT_S16_NE;
            if (ioctl (dsp->fd, SNDCTL_DSP_SETFMT, &dsp->format)){
                log_addf("Can't set fallback audio %d, return %d", tmp, dsp->format);
                
                dsp->format = AFMT_S8;
                if (ioctl (dsp->fd, SNDCTL_DSP_SETFMT, &dsp->format)){
                    log_addf("Can't set fallback audio format %d, return %d", tmp, dsp->format);
                    
                    dsp->format = AFMT_U8;
                    if (ioctl (dsp->fd, SNDCTL_DSP_SETFMT, &dsp->format)){
                        log_addf("Can't set fallback audio format %d, return %d", tmp, dsp->format);
                        goto err;
                    }
                }
            }
        }
        /* fallback ok */
    }
    

    tmp=dsp->channels;
    if (ioctl (dsp->fd, SNDCTL_DSP_CHANNELS, &dsp->channels)){
        log_addf("Can't set channels %d, recommended %d", tmp, dsp->channels);
        
        tmp=dsp->channels;
        if (ioctl (dsp->fd, SNDCTL_DSP_CHANNELS, &dsp->channels)){
            log_addf("Can't set recommended channels %d, return %d", tmp, dsp->channels);
            goto err;
        }
        /* fallback ok */
    }

    tmp=dsp->speed;
    if (ioctl (dsp->fd, SNDCTL_DSP_SPEED, &dsp->speed)){
        log_addf("Can't set samplerate %d, recommended %d", tmp, dsp->speed);
        
        tmp=dsp->speed;
        if (ioctl (dsp->fd, SNDCTL_DSP_SPEED, &dsp->speed)){
            log_addf("Can't set recommended samplerate %d, return %d", tmp, dsp->speed);
            goto err;
        }
        /* fallback ok */
    }

    if (ioctl (dsp->fd, SNDCTL_DSP_GETBLKSIZE, &dsp->blksize)){
        log_addf("Can't set block size %d", tmp, dsp->blksize);
        goto err;
    }   
    goto x;
    
err:;
    if (dsp->fd>=0) close(dsp->fd);
    dsp->fd = -1;

x:;
   /* dbg(" dsp->fd=%d\n", dsp->fd);*/
    return dsp->fd;
} 

int dsp_close(struct dsp *dsp){
    
    ST_START;
    dsp_reset(dsp);
    CONDGFREE(dsp->filename);
    if (dsp->fd >= 0) {
        /*dbg("close_dsp(fd=%d)\n",dsp->fd);*/
        close(dsp->fd);
    }   
    dsp->fd=-1;
    ST_STOP;
    return 0;
}

int dsp_reset(struct dsp *dsp){
    if (!dsp || dsp->fd<0) return 0;
    return ioctl(dsp->fd, SNDCTL_DSP_RESET, NULL); 
}

int dsp_sync(struct dsp *dsp){
    if (!dsp || dsp->fd<0) return 0;
    return ioctl(dsp->fd, SNDCTL_DSP_SYNC, NULL); 
}

int sf2dsp(SF_INFO *sfinfo){
    int format;

    switch (sfinfo->format & SF_FORMAT_SUBMASK){
/*        case SF_FORMAT_PCM_U8:
            format=AFMT_U8;
            break;
        case SF_FORMAT_PCM_S8:
            format=AFMT_S8;
            break;*/
        case SF_FORMAT_ULAW:
            format=AFMT_MU_LAW;
            break;            
        case SF_FORMAT_ALAW:
            format=AFMT_A_LAW;
            break;            
        case SF_FORMAT_IMA_ADPCM:
            format=AFMT_IMA_ADPCM;
            break;            
        default:
            format=AFMT_S16_NE;    
    }
    return format;
}    

int dsp2sf(struct dsp *dsp){
    return 0;
}
